#define _FILE_OFFSET_BITS 64
#include <QtGui>
#include <QTextCodec>
#include "fcrypt.h"
#define COEFF 1024*1024

fcrypt::fcrypt() {
    setupUi(this);

    QTextCodec *codec = QTextCodec::codecForName("UTF-8");
    QTextCodec::setCodecForTr(codec);

    fullSize = 0;

    QStringList header;
    header << tr("Путь") << tr("Размер");
    targetsList->setHorizontalHeaderLabels(header);
    headview = new QHeaderView(Qt::Horizontal, targetsList);
    targetsList->setHorizontalHeader(headview);
    headview->setResizeMode(0, QHeaderView::Interactive);
    headview->setResizeMode(1, QHeaderView::Stretch);
    targetsList->setContextMenuPolicy(Qt::ActionsContextMenu);

    editItemAction = new QAction(tr("Редактировать"), this);
    editItemAction->setEnabled(false);
    deleteItemAction = new QAction(tr("Удалить"), this);
    deleteItemAction->setEnabled(false);

    connect(execButton, SIGNAL(clicked()), this, SLOT(exec()));
    connect(addFile, SIGNAL(clicked()), this, SLOT(addFiles()));
    connect(addDir, SIGNAL(clicked()), this, SLOT(addDirs()));
    connect(editNode, SIGNAL(clicked()), this, SLOT(editItem()));
    connect(deleteNode, SIGNAL(clicked()), this, SLOT(deleteItem()));
    connect(targetsList, SIGNAL(currentCellChanged(int, int, int, int)), this, SLOT(manageFunctions(int, int, int, int)));
    connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));
    connect(aboutAction, SIGNAL(triggered()), this, SLOT(about()));
    connect(aboutQtAction, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
    connect(execAction, SIGNAL(triggered()), this, SLOT(exec()));
    connect(addFileAction, SIGNAL(triggered()), this, SLOT(addFiles()));
    connect(addDirAction, SIGNAL(triggered()), this, SLOT(addDirs()));
    connect(editItemAction, SIGNAL(triggered()), this, SLOT(editItem()));
    connect(deleteItemAction, SIGNAL(triggered()), this, SLOT(deleteItem()));

    targetsList->addAction(editItemAction);
    targetsList->addAction(deleteItemAction);

    QRegExp regExp("[a-zA-Z0-9_]{0,}");
    passLine->setValidator(new QRegExpValidator(regExp, this));
    passConfirmLine->setValidator(new QRegExpValidator(regExp, this));

    statusBar()->showMessage(tr("Объектов: 0; размер: 0 байт"), 0);

    execButton->setIcon(QIcon(":/images/run.png"));
    addFile->setIcon(QIcon(":/images/insert-file.png"));
    addDir->setIcon(QIcon(":/images/insert-directory.png"));
    editNode->setIcon(QIcon(":/images/edit.png"));
    deleteNode->setIcon(QIcon(":/images/remove.png"));
    exitAction->setIcon(QIcon(":/images/exit.png"));
    aboutAction->setIcon(QIcon(":/images/info.png"));
    execAction->setIcon(QIcon(":/images/run.png"));
    addFileAction->setIcon(QIcon(":/images/insert-file.png"));
    addDirAction->setIcon(QIcon(":/images/insert-directory.png"));
    editItemAction->setIcon(QIcon(":/images/edit.png"));
    deleteItemAction->setIcon(QIcon(":/images/remove.png"));
    setWindowIcon(QIcon(":/images/icon.png"));
}

void fcrypt::about() {
    QMessageBox::information(this, tr("О программе File Encryptor"), tr("<h2>File encryptor</h2><br /><b>Версия 1.2</b><br />Простой шифровальщик файлов, основанный на XOR методе. Автор: freesky"));
}

void fcrypt::manageFunctions(int x1, int y1, int x2, int y2) {
    editNode->setEnabled(!(targetsList->currentRow() < 0));
    deleteNode->setEnabled(!(targetsList->currentRow() < 0));
    editItemAction->setEnabled(!(targetsList->currentRow() < 0));
    deleteItemAction->setEnabled(!(targetsList->currentRow() < 0));
}

QString fcrypt::getTextSize(qint64 size) {
    if (size > 1024 * 1024 * 1024)
        return QLocale::system().toString((double)size / (1024 * 1024 * 1024), 'f', 3).append(tr(" Гб"));
    else if (size > 1024 * 1024)
        return QLocale::system().toString((double)size / (1024 * 1024), 'f', 3).append(tr(" Мб"));
    else if (size > 1024)
        return QLocale::system().toString((double)size / 1024, 'f', 3).append(tr(" Кб"));
    else
        return QLocale::system().toString((double)size, 'f', 0).append(tr(" б"));
}


qint64 fcrypt::getSize(QString obj, QString type) {
    if (type == "F")
        return QFileInfo(obj).size();
    else if (type == "D") {
        QStringList objectsList;
        if (!recurseDirs->isChecked()) {
            objectsList = QDir(obj).entryList(QDir::Files | QDir::NoDotAndDotDot | QDir::Hidden | QDir::System);
            for (QStringList::Iterator i = objectsList.begin(); i != objectsList.end(); i++)
                objectsList.append(obj + "/" + *i);
        }
        else
            objectsList = getDirFiles(obj);
    qint64 size = 0;
    for (QStringList::Iterator i = objectsList.begin(); i != objectsList.end(); i++)
        size += QFileInfo(*i).size();
    return size;
    }
    return 0;
}

qint64 fcrypt::getCount() {
    QStringList objectsList;
    objectsList.clear();

    for (int i = 0; i < targetsList->rowCount(); i++) {
        if (targets.at(i) == "F")
            objectsList.append(targetsList->item(i, 0)->text());
        else if (targets.at(i) == "D") {
            if (!recurseDirs->isChecked()) {
                QStringList fDirList = QDir(targetsList->item(i, 0)->text()).entryList(QDir::Files | QDir::NoDotAndDotDot | QDir::Hidden | QDir::System);
                for (QStringList::Iterator fdit = fDirList.begin(); fdit != fDirList.end(); fdit++) {
                    objectsList.append(targetsList->item(i, 0)->text() + "/" + *fdit);
                }
            }
            else
                objectsList = objectsList+getDirFiles(targetsList->item(i, 0)->text());
        }
    }

    return objectsList.size();
}

void fcrypt::updateSBar() {
    statusBar()->showMessage((tr("Объектов: ")+QString::number(getCount())+tr("; размер: ")+getTextSize(fullSize)), 0);  
}

void fcrypt::addFiles() {
    QFileDialog *dlg = new QFileDialog(this);
    QStringList filePaths;
    filePaths = dlg->getOpenFileNames(this, tr("Выберите файлы"), QDesktopServices::storageLocation(QDesktopServices::DocumentsLocation), tr("Все файлы (*.*)"));

    if (filePaths.size() <= 0)
        return;

    for (long i = 0; i < filePaths.size(); i++) {
        targets.append("F");
        targetsList->setRowCount(targetsList->rowCount() + 1);

        QTableWidgetItem *item = new QTableWidgetItem(filePaths.at(i));
        item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
        targetsList->setItem(targetsList->rowCount() - 1, 0, item);

        qint64 size = getSize(filePaths.at(i), "F");
        item = new QTableWidgetItem(getTextSize(size));
        item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
        targetsList->setItem(targetsList->rowCount() - 1, 1, item);
        fullSize += size;
    }

    headview->resizeSections(QHeaderView::ResizeToContents);
    headview->setResizeMode(1, QHeaderView::Stretch);

    updateSBar();

    execButton->setEnabled(!(targetsList->rowCount() == 0));
    execAction->setEnabled(!(targetsList->rowCount() == 0));
}


void fcrypt::addDirs() {
    QFileDialog *dlg = new QFileDialog(this);
    QString dirPath;
    dirPath = dlg->getExistingDirectory(this, tr("Выберите каталог"), QDesktopServices::storageLocation(QDesktopServices::DocumentsLocation));

    if (dirPath.isEmpty())
        return;

    targets.append("D");
    targetsList->setRowCount(targetsList->rowCount() + 1);

    QTableWidgetItem *item = new QTableWidgetItem(dirPath);
    item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
    targetsList->setItem(targetsList->rowCount() - 1, 0, item);
    
    qint64 size = getSize(dirPath, "D");
    item = new QTableWidgetItem(getTextSize(size));
    item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
    targetsList->setItem(targetsList->rowCount() - 1, 1, item);
    fullSize += size;

    headview->resizeSections(QHeaderView::ResizeToContents);
    headview->setResizeMode(1, QHeaderView::Stretch);

    updateSBar();  

    execButton->setEnabled(!(targetsList->rowCount() == 0));
    execAction->setEnabled(!(targetsList->rowCount() == 0));
}

void fcrypt::editItem() {
    if (targetsList->currentRow() < 0)
        return;

    if (targets.at(targetsList->currentRow()) == "F") {
        QFileDialog *dlg = new QFileDialog(this);
        QString filePath;
        filePath = dlg->getOpenFileName(this, tr("Выберите файл"), QDesktopServices::storageLocation(QDesktopServices::DocumentsLocation), tr("Все файлы (*.*)"));

        if (filePath.isEmpty())
            return;

        fullSize -= getSize(targetsList->item(targetsList->currentRow(), 0)->text(), "F");
        QTableWidgetItem *item = new QTableWidgetItem(filePath);
        item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
        targetsList->setItem(targetsList->currentRow(), 0, item);

        qint64 size = getSize(filePath, "F");
        item = new QTableWidgetItem(getTextSize(size));
        item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
        targetsList->setItem(targetsList->currentRow(), 1, item);
        fullSize += size;
        
        headview->resizeSections(QHeaderView::ResizeToContents);
        headview->setResizeMode(1, QHeaderView::Stretch);
    }
    else {
        QFileDialog *dlg = new QFileDialog(this);
        QString dirPath;
        dirPath = dlg->getExistingDirectory(this, tr("Выберите каталог"), QDesktopServices::storageLocation(QDesktopServices::DocumentsLocation));

        if (dirPath.isEmpty())
            return;

        fullSize -= getSize(targetsList->item(targetsList->currentRow(), 0)->text(), "D");
        QTableWidgetItem *item = new QTableWidgetItem(dirPath);
        item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
        targetsList->setItem(targetsList->currentRow(), 0, item);

        qint64 size = getSize(dirPath, "D");
        item = new QTableWidgetItem(getTextSize(size));
        item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
        targetsList->setItem(targetsList->currentRow(), 1, item);
        fullSize += size;
       
        headview->resizeSections(QHeaderView::ResizeToContents);
        headview->setResizeMode(1, QHeaderView::Stretch);
    }

    updateSBar(); 
}

void fcrypt::deleteItem() {
    if (targetsList->currentRow() < 0)
        return;

    if (targets.at(targetsList->currentRow()) == "F")
        fullSize -= getSize(targetsList->item(targetsList->currentRow(), 0)->text(), "F");
    else if (targets.at(targetsList->currentRow()) == "D")
        fullSize -= getSize(targetsList->item(targetsList->currentRow(), 0)->text(), "D");

    targets.removeAt(targetsList->currentRow());
    targetsList->removeRow(targetsList->currentRow());
    targetsList->setCurrentCell(-1, 0);

    updateSBar(); 

    execButton->setEnabled(!(targetsList->rowCount() == 0));
    execAction->setEnabled(!(targetsList->rowCount() == 0));
}


QStringList fcrypt::getDirFiles(const QString& dirName) {
    QDir dir(dirName);
    QStringList fileNames;
    QStringList fileList = dir.entryList(QDir::Files | QDir::NoDotAndDotDot | QDir::Hidden | QDir::System);
    for (QStringList::Iterator fit = fileList.begin(); fit != fileList.end(); fit++)
        fileNames.append(dir.absolutePath() + "/" + *fit);
    QStringList dirList = dir.entryList(QDir::Dirs | QDir::NoDotAndDotDot);
    for (QStringList::Iterator dit = dirList.begin(); dit != dirList.end(); dit++) {
        QDir curDir = dir;
        curDir.cd(*dit);
        QStringList curList = getDirFiles(curDir.absolutePath());
        for (QStringList::Iterator it = curList.begin(); it != curList.end(); it++)
            fileNames.append(QFileInfo(*it).absoluteFilePath());
    }
    return fileNames;
}
    
void fcrypt::exec() {
    if (passLine->text().isEmpty()) {
        QMessageBox::critical(this, tr("Ошибка"), tr("Не введён пароль"));
        return;
    }
    if (passConfirmLine->text().isEmpty()) {
        QMessageBox::critical(this, tr("Ошибка"), tr("Не введено подтверждение пароля"));
        return;
    }
    if (!(passLine->text() == passConfirmLine->text())) {
        QMessageBox::critical(this, tr("Ошибка"), tr("Пароли не совпадают"));
        return;
    }

    QStringList fileList; 
    fileList.clear();

    for (int i = 0; i < targetsList->rowCount(); i++) {
        if (targets.at(i) == "F")
            fileList.append(targetsList->item(i, 0)->text());
        else if (targets.at(i) == "D") {
            if (!recurseDirs->isChecked()) {
                QStringList fDirList = QDir(targetsList->item(i, 0)->text()).entryList(QDir::Files | QDir::NoDotAndDotDot | QDir::Hidden | QDir::System);
                for (QStringList::Iterator fdit = fDirList.begin(); fdit != fDirList.end(); fdit++) {
                    fileList.append(targetsList->item(i, 0)->text() + "/" + *fdit);
                }
            }
            else
                fileList = fileList+getDirFiles(targetsList->item(i, 0)->text());
        }
    }

    progressFullBar->reset();
    progressFullBar->setRange(0, fileList.size());
    progressFullBar->setValue(0);

    for (QStringList::Iterator fit = fileList.begin(); fit != fileList.end(); fit++) {
        QTime mdn(0, 0, 0);
        qsrand(mdn.secsTo(QTime::currentTime()));
        QFile f(*fit), g(*fit+".tmp"+QString::number(qrand() % 65535));

        if (!QFileInfo(f).isFile() || !f.open(QIODevice::ReadOnly) || !g.open(QIODevice::WriteOnly)) {
            progressFullBar->setValue(progressFullBar->value() + 1);
            f.close();
            break;
        }

        char buf[bufferSize->value() * COEFF];
        qint64 l;

        long pI = 0;
        if (evaluateCrypt->isChecked()) {
            for (int i = 0; i < passLine->text().size(); i++)
                pI += (int)passLine->text().at(i).toAscii() * (i + 1);
        }
    
        progressFileBar->reset();
        progressFileBar->setRange(1, f.size());

        while ((l = f.read(buf, sizeof(buf))) > 0) {
            for (int i = 0; i < (bufferSize->value() * COEFF); i++) {
                if (evaluateCrypt->isChecked())
                    *(buf + i) = *(buf + i) ^ pI;
                else
                    for (int j = 0; j < passLine->text().size(); j++)
                        *(buf + i) = *(buf + i) ^ ((int)passLine->text().at(j).toAscii() * (j + 1));
            }
            g.write(buf, l);
            progressFileBar->setValue(progressFileBar->value() + l);
            qApp->processEvents(QEventLoop::ExcludeUserInputEvents);
        }

        f.close();
        g.close();
        f.remove();
        g.rename(*fit);

        progressFullBar->setValue(progressFullBar->value() + 1);
        qApp->processEvents(QEventLoop::ExcludeUserInputEvents);
    }

    QMessageBox::information(this, tr("Информация"), tr("Операция завершена"));
}
